Attribute VB_Name = "InBlu_Math"
'=========================================================================================
' Descrizione.....: Routine for matricial calculations.
' Nome dei Files..: Inblu_Math.bas
' Data............: 12/1/2001
' Versione........: 1.01 , 32 bits.
' Sistema.........: VB6 , Windows 2000.
' Scritto da......: Condor trade s.r.l.
' E-Mail..........: info@condorgroup.com
' web Site........: http://www.inblu.com
'========================================================================================
'   Functions and subroutines:
'
'    c() = MSom(a(), b()):            returns the sum of two matrices.
'    c() = MDif(a(), b()):            returns the difference of two matrices.
'    c() = MMol(a(), b()):            returns the product of two matrices.
'    c() = MMSc(a(),s):               returns the product of one matrix and a scalar.
'    c() = MIde(N):                   returns an identity matrix having order N.
'    c() = Trasposta(a()):            returns the transposed matrix.
'    B = MEgu(a(), b()):              verifies the equality a() = b().
'    B = Inversa(a(), c()):           calculate the inverse matrix.
'    B = Determinante(a(), det):      calculates the determinant of a square matrix.
'    B = LinEqu(a(), b()):            calculates the solutions of a linear system.
'    GAUS a(), b(), x(), IERROR:      calculates the solutions of a linear system.
'    TRISOL a(), b(), c(), d():       solves a tridiagonal system with Thomas' method.
'    B = Jacobi (a(), d(), v()):      calculates eigenvalues and eigenvectors of a
'                                     square, real and symmetrical matrix.
'    B = AutoValori(a(), wr(), wi()): calculates eigenvalues, real and complex, of a
'                                     square real matrix.
'    POWERM a(), v(), lambda:         calculates the main eigenvalue and the
'                                     corrisponding eigenvector, real, of a square real matrix.
'    POWINV a(), v(), alfa, lambda:   calculates the nearest eigenvalue to alfa and the
'                                     corrisponding eigenvector, real, of a square real matrix.
'
'   Types:  a() as Double, b() as Double, c() as Double, d() as Double
'           v(), wr() as Double, wi() as Double
'           d as Double, det as Double, s as Double, alfa as Double, lambda as Double
'           N as Long, IERROR as Long
'           B as Boolean
'
'   Note:  all matrix and vectors used in these routines start from index 1:
'          e.g. a(1 To N, 1 To N)
'

'Declarations
Public Const PiGreco As Double = 3.14159265358979
Option Explicit


Public Function Ceil(ByVal Number As Double) As Double

    'declarations
    Dim Res As Double
    
    Res = Number Mod 1
    If Number < 1 And Number > 0 Then
        Ceil = 1
    ElseIf Number < 0 And Number > -1 Then
        Ceil = 0
    ElseIf Res = 0 Then
        Ceil = Int(Number)
    Else
        Ceil = Int(Number) + 1
    End If
    
End Function

Private Sub balanc(ByRef A#())
'
'   Given an N by N matrix a() stored in an array, this routine replaces
'   it by a balanced matrix with identical eigenvalues.
'   A symmetric matrix is already balanced and is unaffected
'   by this procedure. The parameter RADIX should be the machine's
'   floating-point radix.
'
    Dim I&, J&, N&, Last&
    Dim C#, f#, g#, r#, s#
    Const RADIX# = 2#, SQRDX# = RADIX * RADIX
'
    N = UBound(A, 1)
'
1
    Last = 1
    For I = 1 To N  ' Calculate row and column norms.
        C = 0#
        r = 0#
        For J = 1 To N
            If (J <> I) Then
                C = C + Abs(A(J, I))
                r = r + Abs(A(I, J))
            End If
        Next J
        If (C <> 0# And r <> 0#) Then ' If both are nonzero,
            g = r / RADIX
            f = 1#
            s = C + r
'
2
            If (C < g) Then ' find the integer power of the machine radix that
                            ' comes closest to balancing the matrix.
                f = f * RADIX
                C = C * SQRDX
                GoTo 2
            End If
            g = r * RADIX
'
3
            If (C > g) Then
                f = f / RADIX
                C = C / SQRDX
                GoTo 3
            End If
'
            If ((C + r) / f < 0.95 * s) Then
                Last = 0
                g = 1# / f
                For J = 1 To N ' Apply similarity transformation.
                    A(I, J) = A(I, J) * g
                Next J
                For J = 1 To N
                    A(J, I) = A(J, I) * f
                Next J
            End If
        End If
    Next I
    If (Last = 0) Then GoTo 1
'
'
'
End Sub


Public Sub TRISOL(A#(), ByRef B#(), C#(), ByRef d#())
'
'   Solves a tridiagonal system with Thomas' method.
'    a(1 To N): low diagonal.
'    b(1 To N): middle diagonal.
'    c(1 To N): high diagonal.
'    d(1 To N): enters with known term vector;
'               exits with solutions vector.
'   Has to be:
'    a(1) = 0.
'    c(N) = 0.
'
    Dim I&, J&, N&, r#
'
    N = UBound(A)
'
    For I = 2 To N
        r = A(I) / B(I - 1)
        B(I) = B(I) - r * C(I - 1)
        d(I) = d(I) - r * d(I - 1)
    Next I
'
    d(N) = d(N) / B(N)
    For I = 2 To N
        J = N - I + 1
        d(J) = (d(J) - C(J) * d(J + 1)) / B(J)
    Next I
'
'
'
End Sub


Public Function Jacobi(ByRef A#(), ByRef d#(), ByRef v#() _
    , Optional ByRef NRot&) As Boolean
'
'   Computes all eigenvalues and eigenvectors of a real symmetric matrix a(),
'   which is of size N by N. Only the elements of a() above the diagonal are cosidered.
'   On output, elements of a() above the diagonal are destroyed.
'   d() returns the eigenvalues of a() in its first N elements.
'   v() is a matrix with the same logical and physical dimensions as a(),
'   whose columns contain, on output, the normalized eigenvectors of a().
'   NRot returns the number of Jacobi rotations that were required.
'
    Dim I&, IP&, IQ&, J&, N&
    Dim C#, g#, H#, s#, sm#, T#, tau#, theta#, tresh#
'
    N = UBound(A, 1)
    ReDim d(1 To N), v(1 To N, 1 To N)
    ReDim B#(1 To N), Z#(1 To N)
'
    For IP = 1 To N         ' Initialize to the identity matrix.
        For IQ = 1 To N
            v(IP, IQ) = 0#
        Next IQ
        v(IP, IP) = 1#
    Next IP
'
    For IP = 1 To N
        B(IP) = A(IP, IP)   ' Initialize b and d to the diagonal of a.
        d(IP) = B(IP)
        Z(IP) = 0#          ' This vector will accumulate terms of the form TApq
                            ' as in equation (11.1.14).
    Next IP
'
    NRot = 0
    For I = 1 To 50
        sm = 0#
        For IP = 1 To N - 1 ' Sum off-diagonal elements.
            For IQ = IP + 1 To N
                sm = sm + Abs(A(IP, IQ))
            Next IQ
        Next IP
        'If (sm = 0#) Then
        If (Abs(sm) < 1E-20) Then
            Jacobi = True   ' The normal return, which relies on
            Exit Function   ' quadratic convergence to machine underflow.
        End If
'
        If (I < 4) Then
            tresh = 0.2 * sm / (N ^ 2)  ' ...on the first three sweeps.
        Else
            tresh = 0#                  ' ...thereafter.
        End If
'
        For IP = 1 To N - 1
            For IQ = IP + 1 To N
                g = 100# * Abs(A(IP, IQ))
                ' After four sweeps, skip the rotation if the
                ' off-diagonal element is small:
                If ((I > 4) And ((Abs(d(IP)) + g) = Abs(d(IP))) _
                And ((Abs(d(IQ)) + g) = Abs(d(IQ)))) Then
                    A(IP, IQ) = 0#
                ElseIf (Abs(A(IP, IQ)) > tresh) Then
                    H = d(IQ) - d(IP)
                    If (Abs(H) + g = Abs(H)) Then
                        T = A(IP, IQ) / H
                    Else
                        theta = 0.5 * H / A(IP, IQ) ' Equation (11.1.10).
                        T = 1# / (Abs(theta) + Sqr(1# + theta ^ 2))
                        If (theta < 0#) Then T = -T
                    End If
                    C = 1# / Sqr(1 + T ^ 2)
                    s = T * C
                    tau = s / (1# + C)
                    H = T * A(IP, IQ)
                    Z(IP) = Z(IP) - H
                    Z(IQ) = Z(IQ) + H
                    d(IP) = d(IP) - H
                    d(IQ) = d(IQ) + H
                    A(IP, IQ) = 0#
                    For J = 1 To IP - 1         ' Case of rotations 1 <= j < p.
                        g = A(J, IP)
                        H = A(J, IQ)
                        A(J, IP) = g - s * (H + g * tau)
                        A(J, IQ) = H + s * (g - H * tau)
                    Next J
                    For J = IP + 1 To IQ - 1    ' Case of rotations p < j < q.
                        g = A(IP, J)
                        H = A(J, IQ)
                        A(IP, J) = g - s * (H + g * tau)
                        A(J, IQ) = H + s * (g - H * tau)
                    Next J
                    For J = IQ + 1 To N         ' Case of rotations q < j <= n.
                        g = A(IP, J)
                        H = A(IQ, J)
                        A(IP, J) = g - s * (H + g * tau)
                        A(IQ, J) = H + s * (g - H * tau)
                    Next J
                    For J = 1 To N
                        g = v(J, IP)
                        H = v(J, IQ)
                        v(J, IP) = g - s * (H + g * tau)
                        v(J, IQ) = H + s * (g - H * tau)
                    Next J
                    NRot = NRot + 1
                End If
            Next IQ
        Next IP
'
        For IP = 1 To N
            B(IP) = B(IP) + Z(IP)
            d(IP) = B(IP)   ' Update d with the sum of tapq,
            Z(IP) = 0#      ' and reinitialize z.
        Next IP
    Next I
'
'
Jacobi_ERR:
    Jacobi = False
    MsgBox LoadResString(393), vbCritical, LoadResString(394)
'   MsgBox "Troppe iterazioni.", vbCritical, " Jacobi"
'
'
'
End Function


Private Function ludcmp(ByRef A#(), ByRef Indx&(), ByRef d#) As Boolean
'
'   Given a matrix a(1 To N, 1 To N), this routine replaces it by
'   the LU decomposition of a rowwise permutation of itself.
'   a() is the input matrix. a() is output, arranged as in equation
'   (2.3.14) above; Indx(1 To N) is an output vector that records the
'   row permutation effected by the partial pivoting;
'   d is output as  1 depending on whether the number of row interchanges
'   was even or odd, respectively. This routine is used in combination
'   with lubksb to solve linear equations or invert a matrix.
'
    Dim I&, IMax&, N&, J&, K&, bSing As Boolean
    Dim aamax#, dum#, sum#, vv#()
    Const tiny# = 1E-20     ' A small number.
'
    N = UBound(A, 1)
    ReDim Indx(1 To N)
    ReDim vv#(1 To N)       ' vv stores the implicit scaling of each row.
    bSing = False
'
    d = 1#                  ' No row interchanges yet.
    For I = 1 To N          ' Loop over rows to get the implicit scaling information.
        aamax = 0#
        For J = 1 To N
            If (Abs(A(I, J)) > aamax) Then aamax = Abs(A(I, J))
        Next J
        If (aamax = 0#) Then
            MsgBox LoadResString(395) & vbNewLine _
                 & LoadResString(396), vbCritical, LoadResString(397)
'           MsgBox "Singular matrix in ludcmp." & vbNewLine _
                 & "No nonzero largest element.", vbCritical, " ludcmp"
            ludcmp = False
            Exit Function
        End If
        vv(I) = 1# / aamax  ' Save the scaling.
    Next I
'
    For J = 1 To N          ' This is the loop over columns of Crout's method.
        For I = 1 To J - 1  ' This is equation (2.3.12) except for i = j.
            sum = A(I, J)
            For K = 1 To I - 1
                sum = sum - A(I, K) * A(K, J)
            Next K
            A(I, J) = sum
        Next I
        aamax = 0#          ' Initialize for the search for largest pivot element.
        For I = J To N      ' This is i = j of equation (2.3.12) and i = j +1 ... N
                            ' of equation (2.3.13).
            sum = A(I, J)
            For K = 1 To J - 1
                sum = sum - A(I, K) * A(K, J)
            Next K
            A(I, J) = sum
            dum = vv(I) * Abs(sum)  ' Figure of merit for the pivot.
            If (dum >= aamax) Then  ' Is it better than the best so far?
                IMax = I
                aamax = dum
            End If
        Next I
'
        If (J <> IMax) Then         ' Do we need to interchange rows?
            For K = 1 To N          ' Yes, do so...
                dum = A(IMax, K)
                A(IMax, K) = A(J, K)
                A(J, K) = dum
            Next K
            d = -d                  ' ...and change the parity of d.
            vv(IMax) = vv(J)        ' Also interchange the scale factor.
        End If
        Indx(J) = IMax
1       If (A(J, J) = 0#) Then      ' If the pivot element is zero the matrix is singular
            A(J, J) = tiny          ' (at least to the precision of the algorithm).
                                    ' For some applications on singular matrices, it is
                                    ' desirable to substitute TINY for zero.
            bSing = True
        End If
'
        If (J <> N) Then            ' Now, finally, divide by the pivot element.
            dum = 1# / A(J, J)
            For I = J + 1 To N
                A(I, J) = A(I, J) * dum
            Next I
        End If
    Next J                          ' Go back for the next column in the reduction.
'
    ludcmp = True
    ludcmp = ludcmp And (Not bSing) ' Opzionale: per avvertire della singolarita'.
                                    ' Vedi commento originale all' istruzione 1
                                    ' e sg.
'
'
'
End Function


Private Sub lubksb(A#(), Indx&(), ByRef B#())
'
'   Solves the set of N linear equations A * X = B.
'   Here a(1 To N, 1 To N) is input, not as the matrix A but rather
'   as its LU decomposition, determined by the routine ludcmp.
'   Indx(1 To N) is input as the permutation vector returned by ludcmp.
'   b(1 To N) is input as the right-hand side vector B, and returns
'   with the solution vector X.
'   a() and Indx() are not modified by this routine and can be left
'   in place for successive calls with different right-hand sides b().
'   This routine takes into account the possibility that b() will begin
'   with many zero elements, so it is efficient for use in matrix inversion.
'
    Dim I&, N&, II&, J&, LL&
    Dim sum#
'
    N = UBound(A, 1)
'
    II = 0              ' When II is set to a positive value, it will become the in-
                        ' dex of the first nonvanishing element of b(). We now do
                        ' the forward substitution, equation (2.3.6). The only new
                        ' wrinkle is to unscramble the permutation as we go.
    For I = 1 To N
        LL = Indx(I)
        sum = B(LL)
        B(LL) = B(I)
        If (II <> 0) Then
            For J = II To I - 1
                sum = sum - A(I, J) * B(J)
            Next J
        ElseIf (sum <> 0#) Then
            II = I      ' A nonzero element was encountered, so from now on we will
                        ' have to do the sums in the loop above.
        End If
        B(I) = sum
    Next I
'
    For I = N To 1 Step -1      ' Now we do the backsubstitution, equation (2.3.7).
        sum = B(I)
        For J = I + 1 To N
            sum = sum - A(I, J) * B(J)
        Next J
        B(I) = sum / A(I, I)    ' Store a component of the solution vector X.
    Next I
'
'
'
End Sub


Public Function Inversa_2(ByRef A() As Double, ByRef C() As Double _
    , Optional ByRef det As Double) As Boolean
'
'   Returns, in c(), the inverse of the square matrix a() and in det the
'   determinant of().
'   The original matrix a() is modified.
'   If, in the calculations, abnormal conditions are found, the function
'   returns False.
'   Note: the used algorithm is not so robust: for big matrices use
'   routine Inversa
'
    Dim I As Long, J As Long, L As Long, K As Long, N As Long
    Dim T As Double, s As Double, d As Double
'
    On Error GoTo Inversa_2_ERR
'
    N = UBound(A, 1)
    ReDim C(1 To N, 1 To N) As Double
'
    For I = 1 To N
        For J = 1 To N
            C(I, J) = 0#
        Next J
        C(I, I) = 1#
    Next I
'
    det = 1#
    For J = 1 To N
        For I = J To N
            If A(I, J) <> 0 Then Exit For
        Next I
        If I > N Then Err.Raise 5, , "La matrice e' singolare."
'
        For L = 1 To N
            T = A(J, L)
            A(J, L) = A(I, L)
            A(I, L) = T
            s = C(J, L)
            C(J, L) = C(I, L)
            C(I, L) = s
        Next L
'
        d = A(J, J)
        det = det * d
        For L = 1 To N
            A(J, L) = A(J, L) / d
            C(J, L) = C(J, L) / d
        Next L
'
        For L = 1 To N
            If J <> L Then
                d = A(L, J)
                For K = 1 To N
                    A(L, K) = A(L, K) - A(J, K) * d
                    C(L, K) = C(L, K) - C(J, K) * d
                Next K
            End If
        Next L
    Next J
'
'
Inversa_2_ERR:
    Inversa_2 = (Err = 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, LoadResString(398)
'       MsgBox Err.Description, vbCritical, " Inversa_2"
    End If
'
'
'
End Function


Public Function Inversa(ByRef A#(), ByRef C#()) As Boolean
'
'   Returns, in c(), the inverse of the square matrix a().
'   The original matrix a() is substituted by his decomposition
'   in triangolar matrix, inferior and superior
'
    Dim I&, J&, N&, Indx&(), d#
'
    N = UBound(A, 1)
    ReDim C(1 To N, 1 To N), B#(1 To N)
'
    For I = 1 To N              ' Set up identity matrix.
        For J = 1 To N
            C(I, J) = 0#
        Next J
        C(I, I) = 1#
    Next I
'
    ' Decompose the matrix just once:
    If ludcmp(A(), Indx(), d) Then
        ' Find inverse by columns:
        For J = 1 To N
            For I = 1 To N
                B(I) = C(I, J)
            Next I
            Call lubksb(A(), Indx(), B())
            For I = 1 To N
                C(I, J) = B(I)
            Next I
        Next J
    '
        Inversa = True
    Else
        Inversa = False
    End If
'
'
'
End Function


Public Function Determinante(ByRef A#(), ByRef det#) As Boolean
'
'   Returns, in det, the determinant of the square matrix a().
'   The original matrix a() is substituted by his decomposition
'   in triangolar matrix, inferior and superior
'
    Dim Indx&(), J&
'
    ' This returns det as  1:
    If ludcmp(A(), Indx(), det) Then
        For J = 1 To UBound(A, 1)
            det = det * A(J, J)
        Next J
    '
        Determinante = True
    Else
        Determinante = False
    End If
'
'
'
End Function


Public Function LinEqu(ByRef A#(), ByRef B#()) As Boolean
'
'   Solves a system of N linear equations in N variables.
'    a(1 To N, 1 To N): matrix of coefficient's variables.
'    b(1 To N):         vector fo knowsn terms.
'   The original matrix a() is substituted by his decomposition
'   in triangolar matrix, inferior and superior
'   Vector b() returns with solutions
'
    Dim Indx&(), d#
'
    If ludcmp(A(), Indx(), d) Then
        Call lubksb(A(), Indx(), B())
'
        LinEqu = True
    Else
        LinEqu = False
    End If
'
'
'
End Function


Private Sub elmhes(ByRef A#())
'
'   Reduction to Hessenberg form by the elimination method.
'   The real, nonsymmetric, N by N matrix a(), is replaced by an upper Hessenberg
'   matrix with identical eigenvalues.
'   Recommended, but not required, is that this routine be preceded by balanc.
'   On output, the Hessenberg matrix is in elements a(i,j) with i <= j + 1.
'   Elements with i > j + 1 are to be thought of as zero, but are returned with
'   random values.
'
    Dim I&, J&, M&, N&
    Dim X#, Y#
'
    N = UBound(A, 1)
'
    For M = 2 To N - 1          ' m is called r + 1 in the text.
        X = 0#
        I = M
        For J = M To N          ' Find the pivot.
            If (Abs(A(J, M - 1)) > Abs(X)) Then
                X = A(J, M - 1)
                I = J
            End If
        Next J
'
        If (I <> M) Then        ' Interchange rows and columns.
            For J = M - 1 To N
                Y = A(I, J)
                A(I, J) = A(M, J)
                A(M, J) = Y
            Next J
            For J = 1 To N
                Y = A(J, I)
                A(J, I) = A(J, M)
                A(J, M) = Y
            Next J
        End If
'
        If (X <> 0#) Then       ' Carry out the elimination.
            For I = M + 1 To N
                Y = A(I, M - 1)
                If (Y <> 0#) Then
                    Y = Y / X
                    A(I, M - 1) = Y
                    For J = M To N
                        A(I, J) = A(I, J) - Y * A(M, J)
                    Next J
                    For J = 1 To N
                        A(J, M) = A(J, M) + Y * A(J, I)
                    Next J
                End If
            Next I
        End If
    Next M
'
'
'
End Sub


Private Function hqr(ByRef A#(), ByRef wr#(), ByRef wi#()) As Boolean
'
'   Finds all eigenvalues of an N by N upper Hessenberg matrix a().
'   On input a() can be exactly as output from elmhes (11.5);
'   on output it is destroyed.
'   The real and imaginary parts of the eigenvalues are returned in
'   wr() and wi(), respectively.
'
    Dim I&, ITS&, J&, K&, L&, M&, N&, Nn&
    Dim anorm#, P#, Q#, r#, s#, T#, u#, v#, w#, X#, Y#, Z#
'
    On Error GoTo hqr_ERR
'
    N = UBound(A, 1)
    ReDim wr(1 To N), wi(1 To N)
'
    anorm = 0#      ' Compute matrix norm for possible use in locating
    For I = 1 To N  ' single small subdiagonal element.
        For J = MAX(I - 1, 1) To N
            anorm = anorm + Abs(A(I, J))
        Next J
    Next I
'
    Nn = N
    T = 0#              ' Gets changed only by an exceptional shift.
'
1   If (Nn >= 0.1) Then ' Begin search for next eigenvalue.
        ITS = 0
2       For L = Nn To 2 Step -1     ' Begin iteration: look for single small sub-
                                    ' diagonal element.
            s = Abs(A(L - 1, L - 1)) + Abs(A(L, L))
            If (s = 0#) Then s = anorm
            If ((Abs(A(L, L - 1)) + s) = s) Then GoTo 3
        Next L
        L = 1
3       X = A(Nn, Nn)
        If (L = Nn) Then            ' One root found.
            wr(Nn) = X + T
            wi(Nn) = 0#
            Nn = Nn - 1
        Else
            Y = A(Nn - 1, Nn - 1)
            w = A(Nn, Nn - 1) * A(Nn - 1, Nn)
            If (L = Nn - 1) Then    ' Two roots found...
                P = 0.5 * (Y - X)
                Q = P ^ 2 + w
                Z = Sqr(Abs(Q))
                X = X + T
                If (Q >= 0#) Then   ' ...a real pair.
                    Z = P + SIGN(Z, P)
                    wr(Nn) = X + Z
                    wr(Nn - 1) = wr(Nn)
                    If (Z <> 0#) Then wr(Nn) = X - w / Z
                    wi(Nn) = 0#
                    wi(Nn - 1) = 0#
                Else                ' ...a complex pair.
                    wr(Nn) = X + P
                    wr(Nn - 1) = wr(Nn)
                    wi(Nn) = Z
                    wi(Nn - 1) = -Z
                End If
                Nn = Nn - 2
'
            Else                    ' No roots found. Continue iteration.
                If (ITS = 30) Then Err.Raise 5, , "Troppe iterazioni"
'
                If (ITS = 10 Or ITS = 20) Then  'Form exceptional shift.
                    T = T + X
                    For I = 1 To Nn
                        A(I, I) = A(I, I) - X
                    Next I
                    s = Abs(A(Nn, Nn - 1)) + Abs(A(Nn - 1, Nn - 2))
                    X = 0.75 * s
                    Y = X
                    w = -0.4375 * s ^ 2
                End If
'
                ITS = ITS + 1
                For M = Nn - 2 To L Step -1 ' Form shift and then look for 2 consecu-
                                            ' tive small subdiagonal elements.
                    Z = A(M, M)
                    r = X - Z
                    s = Y - Z
                    P = (r * s - w) / A(M + 1, M) + A(M, M + 1) 'Equation (11.6.23).
                    Q = A(M + 1, M + 1) - Z - r - s
                    r = A(M + 2, M + 1)
                    s = Abs(P) + Abs(Q) + Abs(r)    ' Scale to prevent over ow or under ow.
                    P = P / s
                    Q = Q / s
                    r = r / s
                    If (M = L) Then GoTo 4
                    u = Abs(A(M, M - 1)) * (Abs(Q) + Abs(r))
                    v = Abs(P) * (Abs(A(M - 1, M - 1)) + Abs(Z) + Abs(A(M + 1, M + 1)))
                    If (u + v = v) Then GoTo 4      ' Equation (11.6.26).
                Next M
'
4               For I = M + 2 To Nn
                    A(I, I - 2) = 0#
                    If (I <> M + 2) Then A(I, I - 3) = 0#
                Next I
                For K = M To Nn - 1                 ' Double QR step on rows l to nn
                                                    ' and columns m to nn.
                    If (K <> M) Then
                        P = A(K, K - 1)             ' Begin setup of Householder vector.
                        Q = A(K + 1, K - 1)
                        r = 0#
                        If (K <> Nn - 1) Then r = A(K + 2, K - 1)
                        X = Abs(P) + Abs(Q) + Abs(r)
                        If (X <> 0#) Then
                            P = P / X               ' Scale to prevent over ow or under ow.
                            Q = Q / X
                            r = r / X
                        End If
                    End If
                    s = SIGN(Sqr(P ^ 2 + Q ^ 2 + r ^ 2), P)
                    If (s <> 0#) Then
                        If (K = M) Then
                            If (L <> M) Then A(K, K - 1) = -A(K, K - 1)
                        Else
                            A(K, K - 1) = -s * X
                        End If
                        P = P + s                   ' Equations (11.6.24).
                        X = P / s
                        Y = Q / s
                        Z = r / s
                        Q = Q / P
                        r = r / P
                        For J = K To Nn             ' Row modification.
                            P = A(K, J) + Q * A(K + 1, J)
                            If (K <> Nn - 1) Then
                                P = P + r * A(K + 2, J)
                                A(K + 2, J) = A(K + 2, J) - P * Z
                            End If
                            A(K + 1, J) = A(K + 1, J) - P * Y
                            A(K, J) = A(K, J) - P * X
                        Next J
                        For I = L To MIN(Nn, K + 3) ' Column modification.
                            P = X * A(I, K) + Y * A(I, K + 1)
                            If (K <> Nn - 1) Then
                                P = P + Z * A(I, K + 2)
                                A(I, K + 2) = A(I, K + 2) - P * r
                            End If
                            A(I, K + 1) = A(I, K + 1) - P * Q
                            A(I, K) = A(I, K) - P
                        Next I
                    End If
                Next K
                GoTo 2 ' ...for next iteration on current eigenvalue.
            End If
        End If
        GoTo 1 ' ...for next eigenvalue.
    End If
'
'
hqr_ERR:
    hqr = (Err = 0)
    If Err <> 0 Then
        MsgBox Err.Description, vbCritical, LoadResString(399)
'       MsgBox Err.Description, vbCritical, " hqr"
    End If
'
'
'
End Function


Private Function MAX(ParamArray vD() As Variant) As Variant
'
'   Implements FORTRAN function MAX(D1, D2, ...DN)
'
    Dim J&, J1&, J2&, vDMax As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMax = vD(J1)
    For J = J1 + 1 To J2
        If vDMax < vD(J) Then vDMax = vD(J)
    Next J
'
    MAX = vDMax
'
'
'
End Function


Private Function MIN(ParamArray vD() As Variant) As Variant
'
'   Implements FORTRAN function MIN(D1, D2, ...DN)
'
    Dim J&, J1&, J2&, vDMin As Variant
'
    J1 = LBound(vD)
    J2 = UBound(vD)
    vDMin = vD(J1)
    For J = J1 + 1 To J2
        If vDMin > vD(J) Then vDMin = vD(J)
    Next J
'
    MIN = vDMin
'
'
'
End Function


Private Function SIGN(ByVal dV#, ByVal dS#) As Double
'
'   Returns the absolute value of v, with the sign of s.
'   Implements FORTRAN function SIGN
'
    If dS < 0 Then
        SIGN = -Abs(dV)
    Else
        SIGN = Abs(dV)
    End If
'
'
'
End Function


Public Function AutoValori(A#(), ByRef wr#(), ByRef wi#()) As Boolean
'
'   Calculates all eigenvalues, real and complex, of the square matrix a().
'   Real part of eigenvalues is returned in wr(), the complex one in wi().
'   If, in the calculations, abnormal conditions are found, the function
'   returns False.
'
    balanc A()
    elmhes A()
'
    AutoValori = hqr(A(), wr(), wi())
'
'
'
End Function


Public Function Trasposta(A#()) As Double()
'
'   Returns the transposed of the a() matrix
'
    Dim I&, J&
    ReDim C#(1 To UBound(A, 2), 1 To UBound(A, 1))
'
    For I = 1 To UBound(A, 1)
        For J = 1 To UBound(A, 2)
            C(J, I) = A(I, J)
        Next J
    Next I
'
    Trasposta = C()
'
'
'
End Function


Public Function MDif(A#(), B#()) As Double()
'
'   Returns the difference between two matrix a() + (-b()).
'   The two matrix have to be of the same order
'
    Dim I&, J&
    ReDim C#(1 To UBound(A, 1), 1 To UBound(A, 2))
'
    On Error GoTo MDif_ERR
    If (UBound(A, 1) <> UBound(B, 1)) Or (UBound(A, 2) <> UBound(B, 2)) Then Err.Raise 5
'
    For I = 1 To UBound(A, 1)
        For J = 1 To UBound(A, 2)
            C(I, J) = A(I, J) - B(I, J)
        Next J
    Next I
'
    MDif = C()
'
'
MDif_ERR:
    If Err <> 0 Then
        MsgBox LoadResString(400), vbCritical, LoadResString(401)
'       MsgBox "Le matrici non sono dello stesso ordine.", vbCritical, " MDif"
    End If
'
'
'
End Function


Public Function MMSc(A#(), ByVal s#) As Double()
'
'   Returns the product of matrix a() and scalar s.
'
    Dim I&, J&
    ReDim C#(1 To UBound(A, 1), 1 To UBound(A, 2))
'
    For I = 1 To UBound(A, 1)
        For J = 1 To UBound(A, 2)
            C(I, J) = s * A(I, J)
        Next J
    Next I
'
    MMSc = C()
'
'
'
End Function


Public Function MEgu(A#(), B#()) As Boolean
'
'   Verifies if matrices a() and b() are equal.
'   The two matrices have to be of the same order.
'
    Dim I&, J&
'
    On Error GoTo MEgu_ERR
    If (UBound(A, 1) <> UBound(B, 1)) Or (UBound(A, 2) <> UBound(B, 2)) Then Err.Raise 5
'
    For I = 1 To UBound(A, 1)
        For J = 1 To UBound(A, 2)
            If A(I, J) <> B(I, J) Then
                MEgu = False
                Exit Function
            End If
        Next J
    Next I
'
'
MEgu_ERR:
    MEgu = (Err = 0)
    If Err <> 0 Then
        MsgBox LoadResString(400), vbCritical, " MEgu"
'       MsgBox "Le matrici non sono dello stesso ordine.", vbCritical, " MEgu"
    End If
'
'
'
End Function


Public Function MSom(A#(), B#()) As Double()
'
'   Returns the sum of two matrices a() + b().
'   The two matrices have to be of the same order.
'
    Dim I&, J&
    ReDim C#(1 To UBound(A, 1), 1 To UBound(A, 2))
'
    On Error GoTo MSom_ERR
    If (UBound(A, 1) <> UBound(B, 1)) Or (UBound(A, 2) <> UBound(B, 2)) Then Err.Raise 5
'
    For I = 1 To UBound(A, 1)
        For J = 1 To UBound(A, 2)
            C(I, J) = A(I, J) + B(I, J)
        Next J
    Next I
'
    MSom = C()
'
'
MSom_ERR:
    If Err <> 0 Then
        MsgBox LoadResString(400), vbCritical, " MSom"
'       MsgBox "Le matrici non sono dello stesso ordine.", vbCritical, " MSom"
    End If
'
'
'
End Function


Public Function MMol(A#(), B#()) As Double()
'
'   Returns the product of two matrix a() * b().
'   Column number of a() has to equal row number of b()
'
    Dim I&, J&, K&
    ReDim C#(1 To UBound(A, 1), 1 To UBound(B, 2))
'
    On Error GoTo MMol_ERR
    If UBound(A, 2) <> UBound(B, 1) Then Err.Raise 5
'
    For I = 1 To UBound(A, 1)
        For J = 1 To UBound(B, 2)
            For K = 1 To UBound(A, 2)
                C(I, J) = C(I, J) + A(I, K) * B(K, J)
            Next K
        Next J
    Next I
'
    MMol = C()
'
'
MMol_ERR:
    If Err <> 0 Then
        MsgBox LoadResString(400), vbCritical, " MMol"
'       MsgBox "Le matrici non hanno ordini compatibili.", vbCritical, " MMol"
    End If
'
'
'
End Function


Public Function MIde(ByVal N&) As Double()
'
'   Returns an identity matrix with order N, with N >= 1:
'
    Dim I&
    ReDim C#(1 To N, 1 To N)
'
    For I = 1 To N
        C(I, I) = 1#
    Next I
'
    MIde = C()
'
'
'
End Function


Public Sub POWINV(A#(), ByRef X#(), ByVal Alpha#, ByRef Lambda# _
    , Optional ByVal Tol# = 0.000001, Optional ByVal MITER& = 100, Optional ByRef ITERM&)
'
'   Subroutine POWINV uses the inverse power method with scaling to
'   estimate Eigenvalues of a matrix A.  The calling program must supply
'   the matrix a(), an initial vector x() of euclidean length 1, a scalar
'   Alpha, a tolerance Tol and an integer MITER = maximum number of
'   iterations desired.  The subroutine returns when the difference of
'   two successive estimates is less than Tol in absolute value or when
'   MITER iteration have been executed.  In the first case, a flag, ITERM,
'   is set to 1 and in the second case ITERM is set to 2.  The approximate
'   Eigenvalue is returned as Lambda and an approximate Eigenvector as x().
'
'   Translaction notes: vector x() is dimensioned in this routine, with initial
'                       values x(1, .., N) = 1/Sqr(N).
'
    Dim I&, J&, N&, ITR&, IERROR&
    Dim Temp#, YScale#, Estold#, Estnew#
'
    N = UBound(A, 1)
    ReDim sa#(1 To N, 1 To N), sx#(1 To N), X#(1 To N), Y#(1 To N)
    For I = 1 To N
        X(I) = 1# / Sqr(CDbl(N))
    Next I
'
    ITR = 1
    For I = 1 To N
        For J = 1 To N
            sa(I, J) = A(I, J)
        Next J
    Next I
'
    ' Calculate the initial Eigenvalue approximation:
    For I = 1 To N
        sx(I) = X(I)
        A(I, I) = A(I, I) - Alpha
    Next I
'
    Call GAUS(A(), X(), Y(), IERROR)
    If IERROR = 2 Then GoTo 10
'
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + Y(I) * sx(I)
        YScale = YScale + Y(I) * Y(I)
    Next I
    YScale = Sqr(YScale)
    Estold = 1# / Temp + Alpha
'
    ' Inverse power method iteration with scaling:
4   ITR = ITR + 1
    For I = 1 To N
        For J = 1 To N
            A(I, J) = sa(I, J)
        Next J
    Next I
    For I = 1 To N
        X(I) = Y(I) / YScale
        sx(I) = X(I)
        A(I, I) = A(I, I) - Alpha
    Next I
'
    Call GAUS(A(), X(), Y(), IERROR)
    If IERROR = 2 Then GoTo 10
'
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + Y(I) * sx(I)
        YScale = YScale + Y(I) * Y(I)
    Next I
    YScale = Sqr(YScale)
    Estnew = 1# / Temp + Alpha
'
    ' Test for termination of the inverse power method iteration:
    If (Abs(Estnew - Estold) <= Tol) Then GoTo 8
    If (ITR >= MITER) Then GoTo 9
    Estold = Estnew
    GoTo 4
'
8   ITERM = 1
    Lambda = Estnew
    For I = 1 To N              ' Aggiunta al programma originale
        X(I) = Y(I) / YScale    ' per ottenere valori di scala
    Next I                      ' inferiore.
    Exit Sub
'
9   ITERM = 2
    Lambda = Estnew
    Exit Sub
'
    ' Set ITERM = 3 if the Gauss elimination step fails:
10  ITERM = 3
'
'
'
End Sub


Public Sub POWERM(A#(), ByRef X#(), ByRef Lambda# _
    , Optional ByVal Tol# = 0.000001, Optional ByVal MITER& = 100, Optional ByRef ITERM&)
'
'   Subroutine POWERM uses the power method with scaling to estimate the
'   dominant Eigenvalue of a matrix A.  The calling program must supply
'   the matrix a(), an initial vector x() of euclidean length 1, a tolerance Tol
'   and an integer MITER = maximum number of power iterations desired.
'   The subroutine returns when the difference of two successive estimates is
'   less than Tol in absolute value or when MITER iteration have been executed.
'   In the first case, a flag, ITERM, is set to 1 and in the second case ITERM
'   is set to 2.  The approximate Eigenvalue is returned as Lambda and an
'   approximate Eigenvector as x().
'
'   Translaction notes: vector x() is dimensioned in this routine, with initial
'                       values x(1, .., N) = 1/Sqr(N).
'
    Dim I&, J&, N&, ITR&, IERROR&
    Dim Temp#, YScale#, Estold#, Estnew#
'
    N = UBound(A, 1)
    ReDim X#(1 To N), Y#(1 To N)
    For I = 1 To N
        X(I) = 1# / Sqr(CDbl(N))
    Next I
'
    ITR = 1
'
    ' Calculate the initial Eigenvalue approximation:
    For I = 1 To N
        Y(I) = 0#
        For J = 1 To N
            Y(I) = Y(I) + A(I, J) * X(J)
        Next J
    Next I
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + Y(I) * X(I)
        YScale = YScale + Y(I) * Y(I)
    Next I
    YScale = Sqr(YScale)
    Estold = Temp
'
    ' Power method iteration with scaling:
3   ITR = ITR + 1
    For I = 1 To N
        X(I) = Y(I) / YScale
    Next I
    For I = 1 To N
        Y(I) = 0#
        For J = 1 To N
            Y(I) = Y(I) + A(I, J) * X(J)
        Next J
    Next I
    Temp = 0#
    YScale = 0#
    For I = 1 To N
        Temp = Temp + Y(I) * X(I)
        YScale = YScale + Y(I) * Y(I)
    Next I
    YScale = Sqr(YScale)
    Estnew = Temp
'
    ' Test for termination of the power method iteration:
    If (Abs(Estnew - Estold) <= Tol) Then GoTo 7
    If (ITR >= MITER) Then GoTo 8
    Estold = Estnew
    GoTo 3
'
7   ITERM = 1
    Lambda = Estnew
    Exit Sub
'
8   ITERM = 2
    Lambda = Estnew
'
'
'
End Sub


Public Sub GAUS(ByRef A#(), ByRef B#(), ByRef X#(), ByRef IERROR&)
'
'   Subroutine GAUS uses Gauss elimination (without pivoting) to solve
'   the system A * X = B.  The calling program must supply the matrix A
'   and the vector B. Arrays a() and b() are destroyed in GAUS.
'   The solution is returned in x() and a flag, IERROR, is set to 1 if a()
'   is non singular and is set to 2 it a() is singular.
'
    Dim I&, J&, K&, N&, IP1&, NM1&, NMK&, NP1&
    Dim Temp#, Q#
'
    N = UBound(A, 1)
    ReDim X#(1 To N)
'
    NM1 = N - 1
    For I = 1 To NM1
        ' Search for non-zero pivot element and interchange rows if necessary.
        ' If non-zero pivot element is found, set IERROR = 2 and return.
        For J = I To N
            If (A(J, I) = 0#) Then GoTo 3
            For K = I To N
                Temp = A(I, K)
                A(I, K) = A(J, K)
                A(J, K) = Temp
            Next K
            Temp = B(I)
            B(I) = B(J)
            B(J) = Temp
            GoTo 4
3
        Next J
        GoTo 8
'
        ' Eliminate the coefficients of X(I) in rows I + 1 ... N:
4       IP1 = I + 1
        For K = IP1 To N
            Q = -A(K, I) / A(I, I)
            A(K, I) = 0#
            B(K) = Q * B(I) + B(K)
            For J = IP1 To N
                A(K, J) = Q * A(I, J) + A(K, J)
            Next J
        Next K
    Next I
    If (A(N, N) = 0#) Then GoTo 8
'
    ' Backsolve the equivalent triangularized system, set IERROR = 1
    ' and return:
    X(N) = B(N) / A(N, N)
    NP1 = N + 1
    For K = 1 To NM1
        Q = 0#
        NMK = N - K
        For J = 1 To K
            Q = Q + A(NMK, NP1 - J) * X(NP1 - J)
        Next J
        X(NMK) = (B(NMK) - Q) / A(NMK, NMK)
    Next K
    IERROR = 1
    Exit Sub
'
8   IERROR = 2
'
'
'
End Sub


Private Function Cosh(ByVal Z As Double) As Double
'
'   Return the hyperbolic cosine of z#:
'
    Cosh = (Exp(Z) + Exp(-Z)) / 2#
'
'
'
End Function


Private Function Sinh(ByVal Z As Double) As Double
'
'   Return the hyperbolic sine of z#:
'
    Sinh = (Exp(Z) - Exp(-Z)) / 2#
'
'
'
End Function


Public Function Atan2(ByVal Y As Double, ByVal X As Double) As Double
'
'   Returns the four quadrant arctangent of the real parts of the elements
'   of X and Y   -pi <= atan2(Y,X) <= pi
'

    If Y >= 0 And X > 0 Then
        Atan2 = Abs(Atn(Y / X))
    ElseIf Y >= 0 And X < 0 Then
        Atan2 = PiGreco - Abs(Atn(Y / X))
    ElseIf Y < 0 And X > 0 Then
        Atan2 = -Abs(Atn(Y / X))
    ElseIf Y < 0 And X < 0 Then
        Atan2 = -PiGreco + Abs(Atn(Y / X))
    ElseIf X = 0 And Y > 0 Then
        Atan2 = PiGreco / 2
    ElseIf X = 0 And Y < 0 Then
        Atan2 = -PiGreco / 2
    End If

End Function
