Attribute VB_Name = "Inblu_ShoesUtils"
'=========================================================================================
' Descrizione.....: Module containing utility functions for shoes applications.
' Nome dei Files..: Inblu_ShoesUtils.bas
' Data............: 22/09/03
' Versione........: 1.01 a 32 bits.
' Sistema.........: VB6 , Windows 2000.
' Scritto da......: Condor trade s.r.l.
' E-Mail..........: info@condorgroup.com
' web Site........: http://www.inblu.com
'========================================================================================

Option Explicit


'Public Sub Grading3D1Step(InitialPoints() As P_Type_3D, FinalPoints() As P_Type_3D, Center As P_Type_3D, _
                    LengthX As Double, DX As Double, LengthY As Double, DY As Double, ZPoints() As P_Type_3D)
'Routine che calcola il grading ad un passo di una dima nello spazio.
'PARAMETRI DI INGRESSO:
'   InitialPoints(): vettore di tipo P_Type_3D contenente i punti (x,y,z) della figura da espandere
'   Center: coordinate (x,y,z) del centro di espansione
'   LengthX: Lunghezza di riferimento asse X dell'espansione
'   DX: Incremento asse X
'   LengthY: Lunghezza di riferimento asse Y dell'espansione
'   DY: Incremento asse Y
'   ZPoints(): vettore di tipo P_Type_3D contenente una scansione approssimativa dell'ultimo numero.
'               Permette di ricavare la quota (z) dei punti delle.

'End Sub


'Public Sub Grading1Step(InitialPoints() As P_Type, FinalPoints() As P_Type, Center As P_Type, _
                    LengthX As Double, DX As Double, LengthY As Double, DY As Double, ZPoints() As P_Type)
'Routine che calcola il grading ad un passo di una dima nel piano.
'PARAMETRI DI INGRESSO:
'   InitialPoints(): vettore di tipo P_Type contenente i punti (x,y) della figura da espandere
'   Center: coordinate (x,y) del centro di espansione
'   LengthX: Lunghezza di riferimento asse X dell'espansione
'   DX: Incremento asse X
'   LengthY: Lunghezza di riferimento asse Y dell'espansione
'   DY: Incremento asse Y

'End Sub

'Public Sub DiffGrading3D(MinPoints() As P_Type_3D, MaxPoints() As P_Type_3D, Min_Number As Integer, Max_Number As Integer, FinalPoints() As P_Type_3D)
''This routine mekes the differential grading procedure starting from two sets of
''points, each one rapresenting a shape in the space.
''The differential grading cosist in the discretization of the distance between
''two corresponding point, the first above the first shape, the second above
''the second shape
'
''INPUT PARAMETERS:
''   MinPoints(): P_Type_3D array containing points of the smaller shape (base
''                of grading)
''   MaxPoints(): P_Type_3D array containing points of the bigger shape
''   MinNumber: foot size of the base of grading
''   MaxNumber: foot size of the maximum number of the grading
''   FinalPoints(): P_Type_3D matrix containing, after the return,
''                  in the (i,j) element, the j(th) point of the i(th)
''                  number of grading
'
'    'Declarations
'    Dim N As Integer, NPoints As Integer
'    Dim J As Long, I As Long
'
'    N = (Max_Number - Min_Number) - 1
'    NPoints = UBound(MinPoints)
'
'    'Calculation of the grading points
'    For J = 1 To NPoints
'        For I = 0 To N
'            FinalPoints(I, J).X = MinPoints(J).X + ((MaxPoints(J).X - MinPoints(J).X) / (N + 1)) * I
'            FinalPoints(I, J).Y = MinPoints(J).Y + ((MaxPoints(J).Y - MinPoints(J).Y) / (N + 1)) * I
'            FinalPoints(I, J).Z = MinPoints(J).Z + ((MaxPoints(J).Z - MinPoints(J).Z) / (N + 1)) * I
'        Next I
'    Next J
'
'End Sub


'Public Sub EstimateExpansionCenter(MinPoints() As P_Type_3D, MaxPoints() As P_Type_3D, Min_Number As Integer, Max_Number As Integer, FinalPoints() As P_Type_3D)
''This routine calculates, given two shapes with the same number of points,
''the theorist expansion center.
''INPUT PARAMETERS:
''   MinPoints(): P_Type_3D array containing points of the smaller shape (base
''                of grading)
''   MaxPoints(): P_Type_3D array containing points of the bigger shape
''   MinNumber: foot size of the base of grading
''   MaxNumber: foot size of the maximum number of the grading
''   FinalPoints(): P_Type_3D matrix containing, after the return,
''                  in the (i,j) element, the j(th) point of the i(th)
''                  number of grading
'
'End Sub


Public Sub C_Spline_3D4S(Pi() As P_Type_3D, Pc() As P_Type_3D)
'   Returns, in the vector Pc(), values of the C-spline curve.
'   Curve is calculated in parametrical mode (0 <= u <= 1) with the
'   value 0 of u corresponding to Pc(0) and the value 1 corresponding
'   to Pc(NPC_1)
'
'   Parameters:
'       Pi(0 to NPI - 1):   Vector of points, datas to interpolate
'       Pc(0 to NPC - 1):   Vector of points, calculated, of the interpolant curve


    Dim NPI_1&, NPC_1&, I&, J&
    Dim u#, ui#, uui#
    Dim cof() As P_Type_3D
    Dim JPrev As Long
    Dim PCount As Long
    Dim PiMod() As P_Type_3D
    Dim NPC As Double
    
    ReDim PiMod(UBound(Pi) + 4)
    NPI_1 = UBound(Pi)      ' Number of points to interpolate - 1
    NPC = UBound(Pc)      ' Number of points on the curve - 1
    NPC_1 = -Int(-(NPC / (NPI_1 + 1) * (NPI_1 + 1 + 3))) - 1
    
    PiMod(0) = Pi(NPI_1 - 1)
    PiMod(1) = Pi(NPI_1)
    
    For I = 0 To NPI_1
        PiMod(I + 2) = Pi(I)
    Next I
    
    PiMod(NPI_1 + 2 + 1) = Pi(0)
    PiMod(NPI_1 + 2 + 2) = Pi(1)
    
    NPI_1 = UBound(PiMod)      ' Number of points to interpolate - 1
    
    Call Find_CCof_3D4S(PiMod(), NPI_1 + 1, cof())
    
    ' Curve always starts from Pi(0) -> u = 0:
    Pc(0).X = Pi(0).X
    Pc(0).Y = Pi(0).Y
    Pc(0).Z = Pi(0).Z
    PCount = 0
    JPrev = 1
    
    For I = 1 To NPC_1 - 1
        
        u = CDbl(I) / CDbl(NPC_1)
        J = Int(u * CDbl(NPI_1)) + 1
        
        If (J > 2) And (J < NPI_1 - 1 + 2) And Not (PCount >= NPC) Then
            
            PCount = PCount + 1
            
            If (J > (NPI_1)) Then J = NPI_1
            
            ui = CDbl(J - 1) / CDbl(NPI_1)
            uui = u - ui
            
            If J = JPrev Then
                
                Pc(PCount).X = cof(4, J).X * uui ^ 3 + cof(3, J).X * uui ^ 2 + cof(2, J).X * uui + cof(1, J).X
                Pc(PCount).Y = cof(4, J).Y * uui ^ 3 + cof(3, J).Y * uui ^ 2 + cof(2, J).Y * uui + cof(1, J).Y
                Pc(PCount).Z = cof(4, J).Z * uui ^ 3 + cof(3, J).Z * uui ^ 2 + cof(2, J).Z * uui + cof(1, J).Z
            
            ElseIf J = JPrev + 2 Then
                
                Pc(PCount).X = cof(4, J).X * uui ^ 3 + cof(3, J).X * uui ^ 2 + cof(2, J).X * uui + cof(1, J).X
                Pc(PCount).Y = cof(4, J).Y * uui ^ 3 + cof(3, J).Y * uui ^ 2 + cof(2, J).Y * uui + cof(1, J).Y
                Pc(PCount).Z = cof(4, J).Z * uui ^ 3 + cof(3, J).Z * uui ^ 2 + cof(2, J).Z * uui + cof(1, J).Z
            
            ElseIf J <> JPrev And Not J = JPrev + 2 Then
                
                Pc(PCount).X = PiMod(J - 1).X
                Pc(PCount).Y = PiMod(J - 1).Y
                Pc(PCount).Z = PiMod(J - 1).Z
                JPrev = J
            
            End If
        
        End If
    
    Next I

    ' Curve always ends on Pi(NPI_1) -> u = 1:
    Pc(PCount).X = Pi(0).X
    Pc(PCount).Y = Pi(0).Y
    Pc(PCount).Z = Pi(0).Z
    
End Sub


Private Sub Find_CCof_3D4S(Pi() As P_Type_3D, ByVal NPI&, cof() As P_Type_3D)
'
'   Find the coefficients of the cubic spline
'   using constant interval parameterization:
'
    Dim I&, H#
'
    ReDim s(1 To NPI) As P_Type_3D, f(1 To NPI) As P_Type_3D, cof(1 To 4, 1 To NPI) As P_Type_3D
    ReDim A(1 To NPI) As Double, B(1 To NPI) As Double, C(1 To NPI) As Double
'
    H = 1# / CDbl(NPI - 1)
    For I = 1 To NPI - 2
        A(I) = 1#
        B(I) = 4#
        C(I) = 1#
    Next I
'
    For I = 1 To NPI - 2
        f(I).X = 6# * (Pi(I + 1).X - 2# * Pi(I).X + Pi(I - 1).X) / H / H
        f(I).Y = 6# * (Pi(I + 1).Y - 2# * Pi(I).Y + Pi(I - 1).Y) / H / H
        f(I).Z = 6# * (Pi(I + 1).Z - 2# * Pi(I).Z + Pi(I - 1).Z) / H / H
    Next I
'
    Call TRIDAG_3D4S(A(), B(), C(), f(), s(), NPI - 2)
    For I = 1 To NPI - 2
        s(NPI - I).X = s(NPI - 1 - I).X
        s(NPI - I).Y = s(NPI - 1 - I).Y
        s(NPI - I).Z = s(NPI - 1 - I).Z
    Next I
'
    s(1).X = 0#
    s(NPI).X = 0#
    s(1).Y = 0#
    s(NPI).Y = 0#
    s(1).Z = 0#
    s(NPI).Z = 0#

    For I = 1 To NPI - 1
        cof(4, I).X = (s(I + 1).X - s(I).X) / 6# / H
        cof(4, I).Y = (s(I + 1).Y - s(I).Y) / 6# / H
        cof(4, I).Z = (s(I + 1).Z - s(I).Z) / 6# / H
        cof(3, I).X = s(I).X / 2#
        cof(3, I).Y = s(I).Y / 2#
        cof(3, I).Z = s(I).Z / 2#
        cof(2, I).X = (Pi(I).X - Pi(I - 1).X) / H - (2# * s(I).X + s(I + 1).X) * H / 6#
        cof(2, I).Y = (Pi(I).Y - Pi(I - 1).Y) / H - (2# * s(I).Y + s(I + 1).Y) * H / 6#
        cof(2, I).Z = (Pi(I).Z - Pi(I - 1).Z) / H - (2# * s(I).Z + s(I + 1).Z) * H / 6#
        cof(1, I).X = Pi(I - 1).X
        cof(1, I).Y = Pi(I - 1).Y
        cof(1, I).Z = Pi(I - 1).Z
    Next I
'
'
'
End Sub


Private Sub TRIDAG_3D4S(A#(), B#(), C#(), f() As P_Type_3D, s() As P_Type_3D, ByVal NPI_2&)
'
'   Solves the tridiagonal linear system of equations:
'
    Dim J&, bet#
    ReDim gam#(1 To NPI_2)
'
    If B(1) = 0 Then Exit Sub
'
    bet = B(1)
    s(1).X = f(1).X / bet
    s(1).Y = f(1).Y / bet
    s(1).Z = f(1).Z / bet
    For J = 2 To NPI_2
        gam(J) = C(J - 1) / bet
        bet = B(J) - A(J) * gam(J)
        If (bet = 0) Then Exit Sub
        s(J).X = (f(J).X - A(J) * s(J - 1).X) / bet
        s(J).Y = (f(J).Y - A(J) * s(J - 1).Y) / bet
        s(J).Z = (f(J).Z - A(J) * s(J - 1).Z) / bet
    Next J
'
    For J = NPI_2 - 1 To 1 Step -1
        s(J).X = s(J).X - gam(J + 1) * s(J + 1).X
        s(J).Y = s(J).Y - gam(J + 1) * s(J + 1).Y
        s(J).Z = s(J).Z - gam(J + 1) * s(J + 1).Z
    Next J
'
'
'
End Sub


Public Sub Find_Tan_Angle(Pi() As P_Type_3D, TanAngle() As Double)
'
'   Given a set of point, for each point except the last, this function calculates
'   the angle of the ortogonal line in the point.
'
    Dim NPI_1&, J&
    Dim cof() As P_Type_3D
'
'
    NPI_1 = UBound(Pi)      ' Number of points
'
    Call Find_CCof_3D(Pi(), NPI_1 + 1, cof())
'
    For J = 0 To NPI_1 - 1
        TanAngle(J) = 180 / PiGreco * Atan2(cof(2, J + 1).Y, cof(2, J + 1).X)
    Next J
    TanAngle(NPI_1) = 0
    
End Sub
