Attribute VB_Name = "Inblu_Splines"
'=======================================================================================
' Descrizione.....: Interpolation routine with 2D and 3D CSplines.
' Nome dei Files..: Inblu_CSplines.bas
' Data............: 20/09/03
' Versione........: 1.0 a 32 bits
' Sistema.........: Visual Basic 6.0 , Windows 2000.
' Scritto da......: Condor trade s.r.l.
' E-Mail..........: info@condorgroup.com
' web Site........: http://www.inblu.com
'=======================================================================================
'
'   Used algorithmes:
'   P. Bourke     -  April 1989
'   D. Cholaseuk  -  8/Dec./1999
'

Option Explicit
'
'Define a type of point that can be drawn
'on video

Public Type P_Type
    X As Double         ' X coordinates
    Y As Double         ' Y Coordinates
End Type

Public Type P_Type_3D
    X As Double         ' X coordinates
    Y As Double         ' Y coordinates
    Z As Double         ' Z coordinates
End Type


Public Sub C_Spline(Pi() As P_Type, Pc() As P_Type)
'
'   Returns, in the vector Pc(), the points of the curve C-Spline.
'   Curve is calculated in parametrical mode (0 <= u <= 1) with the
'   value 0 of u corresponding to Pc(0) and the value 1 corresponding
'   to Pc(NPC_1)
'
'   Parameters:
'       Pi(0 to NPI - 1):   Vector of points, datas to interpolate
'       Pc(0 to NPC - 1):   Vector of points, calculated, of the interpolant curve
'
    Dim NPI_1&, NPC_1&, I&, J&
    Dim u#, ui#, uui#
    Dim cof() As P_Type
'
'
    NPI_1 = UBound(Pi)      ' Number of points to interpolate - 1
    NPC_1 = UBound(Pc)      ' Number of points on the curve - 1
'
    Call Find_CCof(Pi(), NPI_1 + 1, cof())
'
    ' Curve always starts from Pi(0) -> u = 0:
    Pc(0).X = Pi(0).X
    Pc(0).Y = Pi(0).Y
'
    For I = 1 To NPC_1 - 1
        u = CDbl(I) / CDbl(NPC_1)
        J = Int(u * CDbl(NPI_1)) + 1
        If (J > (NPI_1)) Then J = NPI_1
        ui = CDbl(J - 1) / CDbl(NPI_1)
        uui = u - ui
        Pc(I).X = cof(4, J).X * uui ^ 3 + cof(3, J).X * uui ^ 2 + cof(2, J).X * uui + cof(1, J).X
        Pc(I).Y = cof(4, J).Y * uui ^ 3 + cof(3, J).Y * uui ^ 2 + cof(2, J).Y * uui + cof(1, J).Y
    Next I
'
    ' Curve always ends on Pi(NPI_1) -> u = 1:
    Pc(NPC_1).X = Pi(NPI_1).X
    Pc(NPC_1).Y = Pi(NPI_1).Y
'
'
'
End Sub


Public Sub C_Spline_3D(Pi() As P_Type_3D, Pc() As P_Type_3D)
'
'   Returns, in the vector Pc(), values of the C-spline curve.
'   Curve is calculated in parametrical mode (0 <= u <= 1) with the
'   value 0 of u corresponding to Pc(0) and the value 1 corresponding
'   to Pc(NPC_1)
'
'   Parameters:
'       Pi(0 to NPI - 1):   Vector of points, datas to interpolate
'       Pc(0 to NPC - 1):   Vector of points, calculated, of the interpolant curve
'
'
    Dim NPI_1&, NPC_1&, I&, J&
    Dim u#, ui#, uui#
    Dim cof() As P_Type_3D
    Dim JPrev As Long
    
    NPI_1 = UBound(Pi)      ' Number of points to interpolate - 1
    NPC_1 = UBound(Pc)      ' Number of points on the curve - 1

    Call Find_CCof_3D(Pi(), NPI_1 + 1, cof())

    ' Curve always starts from Pi(0) -> u = 0:
    Pc(0).X = Pi(0).X
    Pc(0).Y = Pi(0).Y
    Pc(0).Z = Pi(0).Z
    
    JPrev = 1
    For I = 1 To NPC_1 - 1
        
        u = CDbl(I) / CDbl(NPC_1)
        J = Int(u * CDbl(NPI_1)) + 1
        
        If (J > (NPI_1)) Then J = NPI_1
        
        ui = CDbl(J - 1) / CDbl(NPI_1)
        uui = u - ui
        
        If J = JPrev Then
            
            Pc(I).X = cof(4, J).X * uui ^ 3 + cof(3, J).X * uui ^ 2 + cof(2, J).X * uui + cof(1, J).X
            Pc(I).Y = cof(4, J).Y * uui ^ 3 + cof(3, J).Y * uui ^ 2 + cof(2, J).Y * uui + cof(1, J).Y
            Pc(I).Z = cof(4, J).Z * uui ^ 3 + cof(3, J).Z * uui ^ 2 + cof(2, J).Z * uui + cof(1, J).Z
        
        Else
            
            Pc(I).X = Pi(J - 1).X
            Pc(I).Y = Pi(J - 1).Y
            Pc(I).Z = Pi(J - 1).Z
            JPrev = J
        
        End If
            
    Next I
    
    ' Curve always ends on Pi(NPI_1) -> u = 1:
    Pc(NPC_1).X = Pi(NPI_1).X
    Pc(NPC_1).Y = Pi(NPI_1).Y
    Pc(NPC_1).Z = Pi(NPI_1).Z

End Sub



Private Sub Find_CCof(Pi() As P_Type, ByVal NPI&, cof() As P_Type)
'
'   Find the coefficients of the cubic spline
'   using constant interval parameterization:
'
    Dim I&, H#
'
    ReDim s(1 To NPI) As P_Type, f(1 To NPI) As P_Type, cof(1 To 4, 1 To NPI) As P_Type
    ReDim A(1 To NPI) As Double, B(1 To NPI) As Double, C(1 To NPI) As Double
'
    H = 1# / CDbl(NPI - 1)
    For I = 1 To NPI - 2
        A(I) = 1#
        B(I) = 4#
        C(I) = 1#
    Next I
'
    For I = 1 To NPI - 2
        f(I).X = 6# * (Pi(I + 1).X - 2# * Pi(I).X + Pi(I - 1).X) / H / H
        f(I).Y = 6# * (Pi(I + 1).Y - 2# * Pi(I).Y + Pi(I - 1).Y) / H / H
    Next I
'
    Call TRIDAG(A(), B(), C(), f(), s(), NPI - 2)
    For I = 1 To NPI - 2
        s(NPI - I).X = s(NPI - 1 - I).X
        s(NPI - I).Y = s(NPI - 1 - I).Y
    Next I
'
    s(1).X = 0#
    s(NPI).X = 0#
    s(1).Y = 0#
    s(NPI).Y = 0#
    For I = 1 To NPI - 1
        cof(4, I).X = (s(I + 1).X - s(I).X) / 6# / H
        cof(4, I).Y = (s(I + 1).Y - s(I).Y) / 6# / H
        cof(3, I).X = s(I).X / 2#
        cof(3, I).Y = s(I).Y / 2#
        cof(2, I).X = (Pi(I).X - Pi(I - 1).X) / H - (2# * s(I).X + s(I + 1).X) * H / 6#
        cof(2, I).Y = (Pi(I).Y - Pi(I - 1).Y) / H - (2# * s(I).Y + s(I + 1).Y) * H / 6#
        cof(1, I).X = Pi(I - 1).X
        cof(1, I).Y = Pi(I - 1).Y
    Next I
'
'
'
End Sub


Public Sub Find_CCof_3D(Pi() As P_Type_3D, ByVal NPI&, cof() As P_Type_3D)
'
'   Find the coefficients of the cubic spline
'   using constant interval parameterization:
'
    Dim I&, H#
'
    ReDim s(1 To NPI) As P_Type_3D, f(1 To NPI) As P_Type_3D, cof(1 To 4, 1 To NPI) As P_Type_3D
    ReDim A(1 To NPI) As Double, B(1 To NPI) As Double, C(1 To NPI) As Double
'
    H = 1# / CDbl(NPI - 1)
    For I = 1 To NPI - 2
        A(I) = 1#
        B(I) = 4#
        C(I) = 1#
    Next I
'
    For I = 1 To NPI - 2
        f(I).X = 6# * (Pi(I + 1).X - 2# * Pi(I).X + Pi(I - 1).X) / H / H
        f(I).Y = 6# * (Pi(I + 1).Y - 2# * Pi(I).Y + Pi(I - 1).Y) / H / H
        f(I).Z = 6# * (Pi(I + 1).Z - 2# * Pi(I).Z + Pi(I - 1).Z) / H / H
    Next I
'
    Call TRIDAG_3D(A(), B(), C(), f(), s(), NPI - 2)
    For I = 1 To NPI - 2
        s(NPI - I).X = s(NPI - 1 - I).X
        s(NPI - I).Y = s(NPI - 1 - I).Y
        s(NPI - I).Z = s(NPI - 1 - I).Z
    Next I
'
    s(1).X = 0#
    s(NPI).X = 0#
    s(1).Y = 0#
    s(NPI).Y = 0#
    s(1).Z = 0#
    s(NPI).Z = 0#

    For I = 1 To NPI - 1
        cof(4, I).X = (s(I + 1).X - s(I).X) / 6# / H
        cof(4, I).Y = (s(I + 1).Y - s(I).Y) / 6# / H
        cof(4, I).Z = (s(I + 1).Z - s(I).Z) / 6# / H
        cof(3, I).X = s(I).X / 2#
        cof(3, I).Y = s(I).Y / 2#
        cof(3, I).Z = s(I).Z / 2#
        cof(2, I).X = (Pi(I).X - Pi(I - 1).X) / H - (2# * s(I).X + s(I + 1).X) * H / 6#
        cof(2, I).Y = (Pi(I).Y - Pi(I - 1).Y) / H - (2# * s(I).Y + s(I + 1).Y) * H / 6#
        cof(2, I).Z = (Pi(I).Z - Pi(I - 1).Z) / H - (2# * s(I).Z + s(I + 1).Z) * H / 6#
        cof(1, I).X = Pi(I - 1).X
        cof(1, I).Y = Pi(I - 1).Y
        cof(1, I).Z = Pi(I - 1).Z
    Next I
'
'
'
End Sub


Private Sub TRIDAG(A#(), B#(), C#(), f() As P_Type, s() As P_Type, ByVal NPI_2&)
'
'   Solves the tridiagonal linear system of equations:
'
    Dim J&, bet#
    ReDim gam#(1 To NPI_2)
'
    If B(1) = 0 Then Exit Sub
'
    bet = B(1)
    s(1).X = f(1).X / bet
    s(1).Y = f(1).Y / bet
    For J = 2 To NPI_2
        gam(J) = C(J - 1) / bet
        bet = B(J) - A(J) * gam(J)
        If (bet = 0) Then Exit Sub
        s(J).X = (f(J).X - A(J) * s(J - 1).X) / bet
        s(J).Y = (f(J).Y - A(J) * s(J - 1).Y) / bet
    Next J
'
    For J = NPI_2 - 1 To 1 Step -1
        s(J).X = s(J).X - gam(J + 1) * s(J + 1).X
        s(J).Y = s(J).Y - gam(J + 1) * s(J + 1).Y
    Next J
'
'
'
End Sub


Private Sub TRIDAG_3D(A#(), B#(), C#(), f() As P_Type_3D, s() As P_Type_3D, ByVal NPI_2&)
'
'   Solves the tridiagonal linear system of equations:
'
    Dim J&, bet#
    ReDim gam#(1 To NPI_2)
'
    If B(1) = 0 Then Exit Sub
'
    bet = B(1)
    s(1).X = f(1).X / bet
    s(1).Y = f(1).Y / bet
    s(1).Z = f(1).Z / bet
    For J = 2 To NPI_2
        gam(J) = C(J - 1) / bet
        bet = B(J) - A(J) * gam(J)
        If (bet = 0) Then Exit Sub
        s(J).X = (f(J).X - A(J) * s(J - 1).X) / bet
        s(J).Y = (f(J).Y - A(J) * s(J - 1).Y) / bet
        s(J).Z = (f(J).Z - A(J) * s(J - 1).Z) / bet
    Next J
'
    For J = NPI_2 - 1 To 1 Step -1
        s(J).X = s(J).X - gam(J + 1) * s(J + 1).X
        s(J).Y = s(J).Y - gam(J + 1) * s(J + 1).Y
        s(J).Z = s(J).Z - gam(J + 1) * s(J + 1).Z
    Next J
'
'
'
End Sub


Public Sub InterpolateLinear(FirstDimension() As Double, SecondDimension() As Double, FirstDimensionInterpolated() As Double, _
    SecondDimensionInterpolated() As Double)
    
    Dim I As Long
    Dim DeltaFirst As Double, StepFirst As Double
    Dim DeltaSecond As Double, StepSecond As Double



    DeltaFirst = FirstDimension(1) - FirstDimension(0)
    StepFirst = DeltaFirst / 1000
    ReDim Preserve FirstDimensionInterpolated(0)
    FirstDimensionInterpolated(0) = FirstDimension(0)
    For I = 1 To 1000
        ReDim Preserve FirstDimensionInterpolated(I)
        FirstDimensionInterpolated(I) = FirstDimensionInterpolated(I - 1) + StepFirst
    Next I
        
    DeltaSecond = SecondDimension(1) - SecondDimension(0)
    StepSecond = DeltaSecond / 1000
    ReDim Preserve SecondDimensionInterpolated(0)
    SecondDimensionInterpolated(0) = SecondDimension(0)
    For I = 1 To 1000
        ReDim Preserve SecondDimensionInterpolated(I)
        SecondDimensionInterpolated(I) = SecondDimensionInterpolated(I - 1) + StepSecond
    Next I

End Sub
                                    
Public Sub InterpolateLinear3D4S(Pi() As P_Type_3D, Pc() As P_Type_3D)

    Dim NPI_1&, NPC_1&, I&
    Dim u#, ui#, uui#
    Dim cof() As P_Type_3D
    Dim JPrev As Long, PointPerSegment As Long
    Dim DeltaX As Double, DeltaY As Double, DeltaZ As Double
    Dim StepX As Double, StepY As Double, StepZ As Double
    Dim NPoints As Long, NInt As Long
    Dim Adj As Long, H As Long, J As Long
    Dim Firsttime As Long, TempPI() As P_Type_3D

    TempPI = Pi 'Temporary array
    ReDim Preserve TempPI(UBound(TempPI()) + 1)
    TempPI(UBound(TempPI())) = TempPI(0)
   
    NPoints = UBound(TempPI()) + 1
    NInt = UBound(Pc()) + 1
    Firsttime = 1

    PointPerSegment = Fix(NInt / (NPoints - 1))
    
    Adj = NInt - PointPerSegment * (NPoints - 1)
    
    I = 0   'Original points
    H = 0   'Interpolated points
    J = 0   'Counter for PointPerSegment
    
    Do
        
        If Adj > 0 Then
            
            If J = 0 Then
            
                DeltaX = (TempPI(I + 1).X - TempPI(I).X) / (PointPerSegment + 1)
                DeltaY = (TempPI(I + 1).Y - TempPI(I).Y) / (PointPerSegment + 1)
                DeltaZ = (TempPI(I + 1).Z - TempPI(I).Z) / (PointPerSegment + 1)
                Pc(H).X = TempPI(I).X
                Pc(H).Y = TempPI(I).Y
                Pc(H).Z = TempPI(I).Z
                J = J + 1
                H = H + 1
            
            ElseIf J = PointPerSegment Then
            
                Pc(H).X = TempPI(I).X + DeltaX * J
                Pc(H).Y = TempPI(I).Y + DeltaY * J
                Pc(H).Z = TempPI(I).Z + DeltaZ * J
                I = I + 1
                Adj = Adj - 1
                J = 0
                H = H + 1
            
            Else
                
                Pc(H).X = TempPI(I).X + DeltaX * J
                Pc(H).Y = TempPI(I).Y + DeltaY * J
                Pc(H).Z = TempPI(I).Z + DeltaZ * J
                J = J + 1
                H = H + 1
            
            End If
            
        Else
            
            If J = 0 Then
                
                DeltaX = (TempPI(I + 1).X - TempPI(I).X) / (PointPerSegment)
                DeltaY = (TempPI(I + 1).Y - TempPI(I).Y) / (PointPerSegment)
                DeltaZ = (TempPI(I + 1).Z - TempPI(I).Z) / (PointPerSegment)
                Pc(H).X = TempPI(I).X
                Pc(H).Y = TempPI(I).Y
                Pc(H).Z = TempPI(I).Z
                J = J + 1
                H = H + 1
            
            ElseIf J = PointPerSegment - 1 Then
                
                Pc(H).X = TempPI(I).X + DeltaX * J
                Pc(H).Y = TempPI(I).Y + DeltaY * J
                Pc(H).Z = TempPI(I).Z + DeltaZ * J
                I = I + 1
                H = H + 1
                J = 0
                
            Else
                
                Pc(H).X = TempPI(I).X + DeltaX * J
                Pc(H).Y = TempPI(I).Y + DeltaY * J
                Pc(H).Z = TempPI(I).Z + DeltaZ * J
                J = J + 1
                H = H + 1
            
            End If
        
        End If

    Loop Until H = NInt - 1
    
    Pc(NInt - 1).X = TempPI(NPoints - 1).X
    Pc(NInt - 1).Y = TempPI(NPoints - 1).Y
    Pc(NInt - 1).Z = TempPI(NPoints - 1).Z

End Sub

Public Sub InterpolateLinear3D(Pi() As P_Type_3D, Pc() As P_Type_3D)

    Dim NPI_1&, NPC_1&, I&
    Dim u#, ui#, uui#
    Dim cof() As P_Type_3D
    Dim JPrev As Long, PointPerSegment As Long
    Dim DeltaX As Double, DeltaY As Double, DeltaZ As Double
    Dim StepX As Double, StepY As Double, StepZ As Double
    Dim NPoints As Long, NInt As Long
    Dim Adj As Long, H As Long, J As Long
    Dim Firsttime As Long

    NPoints = UBound(Pi()) + 1
    NInt = UBound(Pc()) + 1
    Firsttime = 1

    PointPerSegment = Fix(NInt / (NPoints - 1))
    
    Adj = NInt - PointPerSegment * (NPoints - 1)
    
    I = 0   'Original points
    H = 0   'Interpolated points
    J = 0   'Counter for PointPerSegment
    
    Do
        
        If Adj > 0 Then
            
            If J = 0 Then
            
                DeltaX = (Pi(I + 1).X - Pi(I).X) / (PointPerSegment + 1)
                DeltaY = (Pi(I + 1).Y - Pi(I).Y) / (PointPerSegment + 1)
                DeltaZ = (Pi(I + 1).Z - Pi(I).Z) / (PointPerSegment + 1)
                Pc(H).X = Pi(I).X
                Pc(H).Y = Pi(I).Y
                Pc(H).Z = Pi(I).Z
                J = J + 1
                H = H + 1
            
            ElseIf J = PointPerSegment Then
            
                Pc(H).X = Pi(I).X + DeltaX * J
                Pc(H).Y = Pi(I).Y + DeltaY * J
                Pc(H).Z = Pi(I).Z + DeltaZ * J
                I = I + 1
                Adj = Adj - 1
                J = 0
                H = H + 1
            
            Else
                
                Pc(H).X = Pi(I).X + DeltaX * J
                Pc(H).Y = Pi(I).Y + DeltaY * J
                Pc(H).Z = Pi(I).Z + DeltaZ * J
                J = J + 1
                H = H + 1
            
            End If
            
        Else
            
            If J = 0 Then
                
                DeltaX = (Pi(I + 1).X - Pi(I).X) / (PointPerSegment)
                DeltaY = (Pi(I + 1).Y - Pi(I).Y) / (PointPerSegment)
                DeltaZ = (Pi(I + 1).Z - Pi(I).Z) / (PointPerSegment)
                Pc(H).X = Pi(I).X
                Pc(H).Y = Pi(I).Y
                Pc(H).Z = Pi(I).Z
                J = J + 1
                H = H + 1
            
            ElseIf J = PointPerSegment - 1 Then
                
                Pc(H).X = Pi(I).X + DeltaX * J
                Pc(H).Y = Pi(I).Y + DeltaY * J
                Pc(H).Z = Pi(I).Z + DeltaZ * J
                I = I + 1
                H = H + 1
                J = 0
                
            Else
                
                Pc(H).X = Pi(I).X + DeltaX * J
                Pc(H).Y = Pi(I).Y + DeltaY * J
                Pc(H).Z = Pi(I).Z + DeltaZ * J
                J = J + 1
                H = H + 1
            
            End If
        
        End If

    Loop Until H = NInt - 1
    
    Pc(NInt - 1).X = Pi(NPoints - 1).X
    Pc(NInt - 1).Y = Pi(NPoints - 1).Y
    Pc(NInt - 1).Z = Pi(NPoints - 1).Z

End Sub


Public Sub C_Spline_3D_ForAxis(Pi() As P_Type_3D, Pc() As P_Type_3D, PreviousPointIndex() As Long)
'
'   Returns, in the vector Pc(), values of the C-spline curve.
'   Curve is calculated in parametrical mode (0 <= u <= 1) with the
'   value 0 of u corresponding to Pc(0) and the value 1 corresponding
'   to Pc(NPC_1)
'
'   Parameters:
'       Pi(0 to NPI - 1):   Vector of points, datas to interpolate
'       Pc(0 to NPC - 1):   Vector of points, calculated, of the interpolant curve
'
'
    Dim NPI_1&, NPC_1&, I&, J&
    Dim u#, ui#, uui#
    Dim cof() As P_Type_3D
    Dim JPrev As Long
'
'
    NPI_1 = UBound(Pi)      ' Number of points to interpolate - 1
    NPC_1 = UBound(Pc)      ' Number of points on the curve - 1
'
    Call Find_CCof_3D(Pi(), NPI_1 + 1, cof())
'
    ' Curve always starts from Pi(0) -> u = 0:
    Pc(0).X = Pi(0).X
    Pc(0).Y = Pi(0).Y
    Pc(0).Z = Pi(0).Z
'
    JPrev = 1
    For I = 1 To NPC_1 - 1
        u = CDbl(I) / CDbl(NPC_1)
        J = Int(u * CDbl(NPI_1)) + 1
        PreviousPointIndex(I) = J - 1
        If (J > (NPI_1)) Then J = NPI_1
        ui = CDbl(J - 1) / CDbl(NPI_1)
        uui = u - ui
        If J = JPrev Then
            Pc(I).X = cof(4, J).X * uui ^ 3 + cof(3, J).X * uui ^ 2 + cof(2, J).X * uui + cof(1, J).X
            Pc(I).Y = cof(4, J).Y * uui ^ 3 + cof(3, J).Y * uui ^ 2 + cof(2, J).Y * uui + cof(1, J).Y
            Pc(I).Z = cof(4, J).Z * uui ^ 3 + cof(3, J).Z * uui ^ 2 + cof(2, J).Z * uui + cof(1, J).Z
        Else
            Pc(I).X = Pi(J - 1).X
            Pc(I).Y = Pi(J - 1).Y
            Pc(I).Z = Pi(J - 1).Z
            JPrev = J
        End If
            
    Next I
'
    ' Curve always ends on Pi(NPI_1) -> u = 1:
    Pc(NPC_1).X = Pi(NPI_1).X
    Pc(NPC_1).Y = Pi(NPI_1).Y
    Pc(NPC_1).Z = Pi(NPI_1).Z
'
'
'
End Sub



